function bM = MPfactorFilter (b, fzeros)
% bM = MPfactor (b [, fzeros])
% Find the min-phase factor of a symmetrical polynomial.
% bM - Min-phase polynomial
%
% b  - Original symmetric polynomial
% fzeros - Optional vector of the locations of double zeros in the
%      frequency response. The elements of fzeros are in normalized 
%      frequency (0 to pi). Frequency zeros at 0 or pi correspond to
%      two roots of the polynomial. Other frequency zeros correspond to
%      four zeros of the polynomial. If fzeros is not supplied, this
%      routine tries to identify them from the roots of the polynomial.
% 
% Find the minimum-phase factor for a symmetric polynomial. Any unit circle
% zeros must be double in order for the filter to be factored into min-
% phase and max-phase terms.

% An alternate strategy is to use side information from the filter design
% routine to help find the roots. The locations of the double-order unit
% circle zeros are known from the extremal locations. Because of the
% finite precision grid used, the extremal locations give the position of
% one of the (almost) double roots. The other is close by, but generally
% not exactly at the same location.
%
% Fix up roots for a symmetric polynomial with double unit circle zeros
% - quad root symmetries for complex (non-unit circle) roots
% - double unit circle roots
% - reciprocal real roots

% $Id: MPfactorFilter.m,v 1.9 2009/06/17 16:54:42 pkabal Exp $

if (nargin < 2)
  fzeros = [];
elseif (size (fzeros, 1) == 1)
  fzeros = fzeros';
end

% Check the input polynomial
if (any (b - fliplr (b)))
  error ('Error: Coefficients are not symmetric');
end

% Find the roots and sort them by magnitude and then angle
% Roots will return a smaller number of roots if the leading coefficients
% of the polynomial are zero - this can be fixed
rt = sort (roots (b));
if (length (rt) ~= length (b) - 1)
  error ('MPfactorFilter: Too few roots returned');
end

% Generate the complex locations of the supplied angular positions
[FrtUR, FrtUC] = RefPosUC (fzeros);

% Remove roots from rt that are close to the supplied zeros
rt = RemoveRoots (rt, [FrtUC; FrtUR]);

% Classify the roots
%  rtIC - complex roots inside the unit circle
%  rtIR - real roots inside the unit circle
%  rtUC - complex unit circle zeros
%  rtUR - real unit circle zeros
%  rtOC - complex roots outside the unit circle
%  rtOR - real roots outside the unit circle
[rtIC, rtIR, rtUC, rtUR, rtOC, rtOR] = SegRoots (rt);

% Check reciprocal complex non-unit circle roots
rtIC = RootIC (rtIC, rtOC);

% Check and fix-up reciprocal real non-unit circle roots
rtIR = RootIR (rtIR, rtOR);

% Pick up half of the roots near z=-1 and z=+1
rtUR1 = RootUR (rtUR);
rtUR1 = [rtUR1 FrtUR];     % Add back the given real zeros

% Pick up half of the complex unit circle roots
rtUC1 = RootUC (rtUC);
rtUC1 = [rtUC1 FrtUC];     % Add back the given zeros

% Find the polynomial corresponding to the min-phase roots
bmin = MProots2Poly ([rtIC; rtUC1; rtIR; rtUR1]);

% Re-normalization
% The original polynomial is the convolution of bmin and bmax, where bmin is
% monic and bmax is the time-reversal of bmin. The required gain can be
% calculated from any of the coefficients of b. Using zero-based indexing and
% letting N be the number of coefficients in bmin and bmax,
%              N-1
%   b[n] = G^2 SUM bmin[k] bmax[n-k] .
%              k=0
% Using bmax[n] = bmin[N-1-n],
%              N-1
%   b[n] = G^2 SUM bmin[k] bmin[N-1-n+k].
%              k=0
% For n=N-1,
%                N-1
%   b[N-1] = G^2 SUM bmin[k]^2.
%                k=0
% This is the largest coefficient in b and will be used to de-normalize the
% minimum-phase factor to give it the correct energy.
%   G = sqrt (b[N-1] / SUM bmin[k]^2).
N = length (bmin);
G = sqrt (b(N) / (bmin * bmin'));
bM = G * bmin;

% Lets pick a reasonable sign for the minimum-phase filter
% - Calculate B0, the frequency response at zero (for LP filters)
% - Calculate Bp, the frequency response at pi (for HP filters)
% - Choose the largest in magnitude and adjust the sign to make
%   that value positive.
B0 = sum (bM);
Bp = sum (bM .* (-1).^(0:N-1));
if (abs (B0) > abs (Bp))
  bM = sign (B0) * bM;
elseif (abs (Bp) > abs (B0))
  bM = sign (Bp) * bM;
end

return

%====================
function [FrtUR, FrtUC] = RefPosUC (fzeros)
% Given the angular locations of zeros on the unit circle, generate
% the complex root values

% Segregate the supplied zeros
I = (sin (fzeros) < eps);           % Real (or near-real) roots
FrtUR = sign (cos (fzeros(I)));
FrtUCu = exp(1j*fzeros(~I));        % Upper half unit circle roots

FrtUC = [FrtUCu; conj(FrtUCu)];

return

%====================
function [rtIC, rtIR, rtUC, rtUR, rtOC, rtOR] = SegRoots (rt)

% Parameters
ImagTol = 1e-4;
UnitTol = 1e-4;

% Segregate roots: inside, unit circle, outside roots
II = (abs (rt) < 1 - UnitTol);
IU = (1 - UnitTol <= abs (rt) & abs (rt) <= 1 + UnitTol);
IO = (1 + UnitTol < abs (rt));

% Segregate inside roots: real, complex
rtI = rt(II);
IR = (abs (imag (rtI)) < sqrt (eps));
rtIR = rtI(IR);
rtIC = rtI(~IR);

% Segregate outside roots: real, complex
rtO = rt(IO);
IR = (abs (imag (rtO)) < sqrt (eps));
rtOR = rtO(IR);
rtOC = rtO(~IR);

% Segregate unit circle roots: Pick up multiple roots near z=1 and z=-1
% that may have split into complex roots
rtU = rt(IU);
IR = (abs (imag(rtU)) < 0.5 * ImagTol);
rtUR = rtU(IR);
rtUC = rtU(~IR);

return

%====================
function rtIC = RootIC (rtIC, rtOC)
% Given the inside roots and the outside roots (only used for
% error checking), do error checking and return the inside roots
% (albeit in a different order).

UnitTol = 1e4 * eps;

% Sort the roots
% - rtIC, sort by imag (and real within imag)
% - rtOC, sort by imag (and negative real within imag)
[xx, II] = sort (real (rtIC));
rtI = rtIC(II);
[xx, IO] = sort (-real (rtOC));
rtO = rtOC(IO);

[xx, II] = sort (imag (rtI));
rtI = rtI(II);
[xx, IO] = sort (imag (rtO));
rtO = rtO(IO);

% Ideally, the roots are (j = N+1-i)
%   rtI(i) = rtI(j)'     % Conjugate symmetry
%   rtO(i) = rtO(j)'     % Conjugate symmetry
%   rtI(i) = 1 / rtO(j)  % reciprocals

NrI = length (rtI);
NrO = length (rtO);
if (NrI ~= NrO)
  error ('Error: Complex non-unit circle roots are not paired');
end
if (mod (NrI, 2) ~= 0)
  error ('Error: Number of complex roots is not even');
end

MaxRError = 0;
rtIC = rtI;
for (i = 1:NrI/2)
  j = NrI + 1 - i;

  if (rtI(i) ~= conj(rtI(j)) && rtO(i) ~= conj(rtO(j)))
    error ('Error: Complex non-unit circle roots are not conjugates');
  end

  RError = abs (abs(rtI(i)) * abs(rtO(i)) - 1);
  MaxRError = max (MaxRError, RError);

% The roots can be surprisingly far from being reciprocals, but forcing
% them to be reciprocals messes up the frequency response.
%  r = abs(rtI(i)) * abs(rtO(i));
%  rtIC(i) = rtI(i) / sqrt (r);

end

if (MaxRError > UnitTol)
  disp ('Warning: Complex non-unit circle roots are not near-reciprocal');
end

return

%====================
function rtIR = RootIR (rtIR, rtOR)

UnitTol = 1e-10;

% Sort the roots
% - rtIR, sort by amplitude, then sign
% - rtOR, sort by reverse amplitude, then sign
rtI = sort (rtIR);
S = (rtI < 0);
rtI = [rtI(S); rtI(~S)];
rtO = sort (rtOR, 'descend');
S = (rtO < 0);
rtO = [rtO(S); rtO(~S)];

NI = length (rtI);
NO = length (rtO);
if (NI ~= NO)
  error ('Error: Real non-unit circle roots are not paired');
end

for (i = 1:NI)
  if (abs (rtI(i) * rtO(i) - 1) > UnitTol)
    error ('Error: Real non-unit circle roots are not near reciprocals');
  end

% Converge the roots
  r = rtI(i) * rtO(i);
  rtIR(i) = rtI(i) / sqrt (r);

end

return

%====================
function rtUR1 = RootUR (rtUR)

% Roots near -1 and +1 are must be double
rtP = rtUR (real (rtUR) > 0);
rtN = rtUR (real (rtUR) < 0);
NP = length (rtP);
NN = length (rtN);
if (mod (NP, 2) ~= 0)
  error ('Error: Unit circle roots are not double (z=+1)');
end
if (mod (NN, 2) ~= 0)
  error ('Error: Unit circle roots are not double (z=-1)');
end

rtUR1 = [];
for (i = 1:NP/2)
  rtUR1 = [rtUR1; +1];
end
for (i = 1:NN/2)
  rtUR1 = [rtUR1; -1];
end

return

%====================
function rtUC1 = RootUC (rtUC)
% Fix up double unit circle roots

% ATAN2C (c)
ATAN2C = @(c) (atan2(imag(c), real(c)));

AngTol = 0.01;

% Sort the unit circle roots by the angle
[xD, I] = sort (ATAN2C(rtUC));
rtUC = rtUC(I);

% Roots on the unit circle are double
N = length (rtUC);
rtUC1 = rtUC(1:2:N);
rtUC2 = rtUC(2:2:N);

Nh = N / 2;
if (mod (Nh, 2) ~= 0)
  error ('Error: Unit circle roots are not paired');
end

for (i = 1:Nh/2)
  j = Nh - i + 1;

% Checks on the roots
  if (ATAN2C(rtUC2(i)) - ATAN2C(rtUC1(i)) > AngTol)
    error ('Error: Unit circle roots are not double');
  end
  if (rtUC1(j) ~= conj(rtUC2(i)) || rtUC2(j) ~= conj(rtUC1(i)))
    error ('Error: Unit circle zeros are not conjugate');
  end

% Converge the roots and ensure that they lie on the unit circle
  rt = 0.5 * (rtUC1(i) + rtUC2(i));
  rtUC1(i) = rt / abs(rt);
  rtUC1(j) = rtUC1(i)';
end

return

%====================
function rtR = RemoveRoots (rt, FrtU)

% Find the minimum spacing between the given roots
minD = min (abs (diff (FrtU)));

% Remove roots from rt that are close to the given roots
rtR = rt;
for (k = 1:length(FrtU))
  I = abs (rtR - FrtU(k)) < 0.1 * minD;
  if (sum(I) ~= 2)
    error ('Error: No roots near supplied roots');
  end
  rtR = rtR(~I);
end

return
