function p = MProots2Poly (rt)
%  p = Roots2Poly (rt)
%  Form a polynomial from classified roots
%  p   - polynomial (polynomial in z^(-1)). This polynomial is
%        (approximately) monic.
%  rt  - Roots. The roots should be in complex pairs or purely real.
%
% The strategy is to calculate the spectrum and then use an FFT to invert
% the spectrum to get the filter coefficients. Direct convolution to find
% the filter coefficients from the root factors is prone to numerical
% problems. Changing the order of the root factors can give very different
% results. Here we calculate the spectrum. This requires an N term complex
% product for each frequency point. If we can calculate an accurate
% spectrum, the inverse DFT will give filter coefficients which have this
% reproducible frequency response.

% $Id: MProots2Poly.m,v 1.9 2009/06/17 16:54:13 pkabal Exp $

% Numerical considerations:
% - Use a power of 2 for the DFT to reduce the number of computations and
%   hence the accuracy. In the worst case, the length of the DFT has to be
%   almost doubled. For filter lengths greater than about 16, the FFT
%   method is more efficient.
% - Remove the linear phase term. After computing the coefficients, they
%   are rotated to compensate.
% - Segregate different types of roots. Unit circle zeros give a real
%   amplitude (after removing the linear phase term). These terms are
%   gathered and multiplied together before using the terms which give rise
%   to a complex amplitude.
%   - rtUC: Unit circle complex roots. The frequency response is an integer
%       delay term times a purely real amplitude.
%   - rtUR: Unit circle real roots, With an even number of roots, the
%       frequency response is an integer delay term times a purely real or
%       purely imaginary term. With an odd number of roots, the response is
%       complex.
%   - rtC and rtR: Complex roots, real roots: These give complex frequency
%       responses

ExTol = 1e-8;
UnitTol = 1000 * eps;

% Segregate real and complex roots
I = (imag(rt) == 0);
rtR = rt(I);
rtC = rt(~I);

Npoly = length (rt) + 1;

% Use a power of two for the DFT
L2N = nextpow2 (Npoly);
M = 2^L2N;
Mh = M / 2;

% Frequency points
w = 2 * pi * (0:Mh) / M;
z = exp (-1j * w);
if (M >= 2)
  z(M/2+1) = -1;     % Force z at pi to be purely real
end
if (M >= 4)
  z(M/4 + 1) = -1j;  % Force z at pi/2 to be purely imaginary
end

% Frequency response
[GC, dC] = FreqRespC (UpperRoots (rtC), z);
[GR, dR] = FreqRespR (rtR, z);
G = GR .* GC;
G = ConjSym (G, M);

d = dC + dR;

PG = real (ifft (G));
PG = Rotate (PG, -d);

% Check the quality of the inverse DFT
% Desired coefficients are 1:Npoly; the remainder should be zero
Ep = PG(1:Npoly) * PG(1:Npoly)';
if (L2N > Npoly)
  Ex = PG(Npoly+1:end) * PG(Npoly+1:end)';
else
  Ex = 0;
end
if (Ex > ExTol * Ep || abs (PG(1) - 1) > UnitTol)
  disp ('Warning - MProots2Poly: Possible loss of precision');
end

p = PG(1:Npoly);

return

%====================
function rtCu = UpperRoots (rtC)

% Complex conjugate root pairs
% Find upper half plane roots
Nr = length (rtC);
if (mod (Nr, 2) ~= 0)
  error ('Error: Roots must be paired');
end
[xx, II] = sort (imag(rtC));
rtC = rtC(II);
rtCu = rtC(1:Nr/2);
rtCl = rtC(Nr:-1:Nr/2+1);
if (rtCu ~= conj(rtCl))
  error ('Error: Roots not in conjugate pairs');
end

return

%====================
function [G, d] = FreqRespC (rtCh, z)

% Given complex roots, find the frequency response. The given roots are
% half of the roots, with the other roots being the complex conjugates of
% the given roots.

% The output is the complex frequency response factored as the product of
% a linear phase term corresponding to an integer delay and a complex
% amplitude,
%   Gx(k) = z(k) * G(k),
% where z(k) = exp (-j w(k)). For roots on the unit circle, the amplitude
% is real.

% Root factor is (1 - z_i z(k))(1 - z_i' z(k))
% For a complex root at z_i, the response is
%   G_i(z(k)) = z(k) A_i(z(k)),
% where
%   A_i(w) = (1+|z_i|^2) cos(w) - 2 Re(z_i) + j (1-|z_i|^2) sin(w)
zR2 = 2 * real (rtCh);
z2 = real(rtCh).^2 + imag(rtCh).^2;
zP = 1 + z2;
zM = 1 - z2;
I = (abs (z2 - 1) < 5 * eps);
zP(I) = 2;         % For unit circle zeros
zM(I) = 0;         % For unit circle zeros

II = (abs (z2 - 1) < 1e-4);
if (I ~= II)
  warning ('Near unit circle zeros')
end

N = length (z);
c = real (z);
s = -imag (z);
G = zeros (1, N);
for (k = 1:N)
  G(k) = prod ((zP * c(k) - zR2) + 1j * (zM * s(k)));
end

d = length (rtCh);

return

%====================
function [G, d] = FreqRespR (rtR, z)

% Given real roots, find the frequency response. 

% The output is the complex frequency response factored as the product of
% a linear phase term corresponding to an integer delay and a complex
% amplitude,
%   Gx(k) = exp(-j * w(k) * d) * G(k).
  
% Root factor is (1 - z_1 z^(-1))(1 - z_2 z^(-1)) or (1 - z_1 z^(-1))
% For real roots z_1 and z_2, the response is
%   G(w) = exp(j w) A(w),
% where
%   A(w) = (1+z1 z2) cos(w) - (z1 + z2) + j (1-z1 z2) sin(w)
rtR = sort (rtR);     % Pair large magnitude with small magnitude
Nr = length (rtR);
if (mod (Nr, 2) == 0)
  Nrh = Nr / 2;
  rtR1 = rtR(1:Nrh);
  rtR2 = rtR(Nr:-1:Nrh+1);
  rtR0 = [];
else
  Nrh = (Nr - 1) / 2;
  rtR1 = rtR(1:Nrh);
  rtR2 = rtR(Nr:-1:Nrh+2);
  rtR0 = rtR(Nrh+1);
end

z1z2 = rtR1 .* rtR2;
z1Pz2 = rtR1 + rtR2;
zP = 1 + z1z2;
zM = 1 - z1z2;
c = real (z);
s = -imag (z);
N = length (z);
G = zeros (1, N);
for (k = 1:N)
  G(k) = prod (zP * c(k) - z1Pz2 + 1j * (zM * s(k)));
  if (mod (Nr, 2) ~= 0)
    G(k) = G(k) * (1 - z(k) * rtR0);
  end
end

d = Nrh;

return

%====================
function Gs = ConjSym (G, M)

if (mod (M, 2) == 0)
  Mh = M / 2;
else
  Mh = (M + 1) / 2;
end
if (length (G) ~= Mh+1)
  error ('Error: Invalid vector length');
end
  
Gs = [G conj(fliplr(G(2:M-Mh)))];

return

%====================
function y = Rotate (x, d)

if (floor (d) ~= d)
  error ('Error: Rotation value is not an integer');
end

N = length (x);
I = mod ((0:N-1) + d, N) + 1;
y = x(I);

return
